\ ICFP 2002 contest

\ socket interface

\ include random.fs

module player

also DOS also

legacy off

libc gethostbyname int (int) gethostbyname ( name -- hostent )
libc socket int int int (int) socket ( class type proto -- fd )
libc connect int int int (int) connect ( fd sock size -- err )
libc htonl int (int) htonl ( x -- x' )

struct{
    cell h_name
    cell h_aliases
    cell h_addrtype
    cell h_length
    cell h_addr_list
} hostent

struct{
    cell family+port
    cell sin_addr
    cell padding1
    cell padding2
} sockaddr_in

Create sockaddr-tmp
sockaddr-tmp sizeof sockaddr_in dup allot erase

: c-string ( addr u -- addr' )
    tuck pad swap move pad + 0 swap c! pad ;

: host>addr ( addr u -- x )
    \ converts a internet name into a IPv4 address
    \ the resulting address is in network byte order
    c-string gethostbyname dup 0= abort" address not found"
    hostent h_addr_list @ @ @ ;

2 Constant PF_INET
1 Constant SOCK_STREAM
6 Constant IPPROTO_TCP

: open-socket ( addr u port -- fid )
    htonl PF_INET [ base c@ 0= ] [IF] $10 lshift [THEN]
    or sockaddr-tmp sockaddr_in family+port !
    2dup !fid >r
    host>addr sockaddr-tmp sockaddr_in sin_addr !
    PF_INET SOCK_STREAM IPPROTO_TCP socket
    dup 0<= abort" no free socket" >r
    r@ sockaddr-tmp $10 connect 0< abort" can't connect"
    r> r@ filehandle ! $7FFFFFFF r@ filesize ! r> ;

previous previous

\ file descriptor

0 Value fd

\ map vars

0 Value robot
0 Value maxrob

0 Value map
0 Value map'
0 Value map''
0 Value costs

0 Value buf1
0 Value buf2

Variable buf1#
Variable buf2#

Variable packs
Variable curious-packs
Create pack 0 , 0 ,    0 ,  0 ,
          \ id  weight dest pos/robot

\ Positions

0 Value w
0 Value h

: robvar ( -- )  Create 0 , DOES>
    dup @ 0= IF  dup >r s" " r> $!  THEN  $@ drop robot cells + ;

robvar alife
robvar x
robvar y

\ mark packets

: cut-tail ( current end size pack -- )
    >r >r r@ - swap r@ move
    r> r> dup $@len rot - swap $!len ;
: pos ( -- addr ) x @ y @ w * + ;

: mark-robot ( id -- )
    packs $@ bounds ?DO
	dup I @ = IF  robot negate I 3 cells + !  LEAVE  THEN
    4 cells +LOOP drop ;
: mark-pos ( id -- )
    packs $@ bounds ?DO
	dup I @ = IF  pos I 3 cells + !
	    I 2 cells + 2@ = IF  I I' 4 cells packs cut-tail  THEN
	    drop unloop EXIT  THEN
    4 cells +LOOP cr ." Put packet " dup . ." at pos " pos . ." into curious"
    pos sp@ 2 cells curious-packs $+! 2drop ;

: find-pack ( id -- )
    packs $@ bounds ?DO
	dup I @ = IF  I pack 4 cells move  LEAVE  THEN
    4 cells +LOOP  drop ;
: unmark-robot ( id -- )
    packs $@ bounds ?DO
	dup I @ = IF  pos I 3 cells + !  LEAVE  THEN
    4 cells +LOOP  drop ;   

\ robot parameters

0 Value id
0 Value carry
0 Value money

\ Socket read in primitives

: get-line ( addr -- addr u )
    dup 1- BEGIN  1+ dup 1 fd read-file throw  WHILE  dup c@ #lf = UNTIL  THEN
    over - ;

\ Read in a map

: .2, ( addr -- ) dup c@ '. = IF  ', swap c!  ELSE  drop  THEN ;
: annotate-map ( -- )
    map map' w h * move map' w h * bounds ?DO
	I c@ '. = IF
	    I 1- c@ '~ = I 1+ c@ '~ = xor 1 and
	    I 1- c@ '# = I 1+ c@ '# = or invert and
	    I w - c@ '~ = I w + c@ '~ = xor 1 and +
	    I w - c@ '# = I w + c@ '# = or invert and
	    s" .,;" drop + c@ I c!  THEN
	I c@ '. = IF
	    I 1- c@ '# = I 1+ c@ '# = and
	    I w - c@ '# = I w + c@ '# = and or
	    IF  ': I c!  THEN  THEN
    LOOP ;

: read-map ( -- ) decimal
    pad get-line evaluate
    2+ to h 2+ to w
    w h * allocate throw to map
    w h * allocate throw to map'
    w h * allocate throw to map''
    map w h * '# fill
    map w 1+ + w h 2- * bounds ?DO
	I w 2- fd read-file throw drop
	pad 1 fd read-file throw drop
    w +LOOP
    w h * cells allocate throw to costs
    w h max 4* cells dup allocate throw to buf1
    allocate throw to buf2 annotate-map ;

: .map ( -- )
    map w h * bounds ?DO
    cr I w type w +LOOP ;
: .map' ( -- )
    map' w h * bounds ?DO
    cr I w type w +LOOP ;
: .map'' ( -- )
    map'' w h * bounds ?DO
    cr I w type w +LOOP ;

: read-par ( -- )
    pad get-line evaluate  to money to carry to id ;

\ Robot vocabulary

Vocabulary robot-voc

: change-len ( addr oldlen -- )
    >r dup maxrob cells swap $!len $@ r> /string erase ;

: expand ( n -- )  1+ dup maxrob <= IF  drop EXIT  THEN
    maxrob cells >r to maxrob
    & alife r@ change-len
    & x     r@ change-len
    & y     r> change-len ;

: #rob ( addr -- ) >r
    r@ 1+ c@ '# = IF
	r> count 1 /string s>number drop dup expand
	state @ IF
	    postpone literal & robot postpone ALiteral postpone !
	ELSE  to robot alife on  THEN
	EXIT  THEN
    r> defers notfound ;

' #rob is notfound

: .rob ( -- ) ." X "  x ? ." Y " y ? ;

: get# ( -- n )  bl word number drop ;

Variable my-p
Variable other-p

: delete-p ( n addr -- )
    dup $@ bounds ?DO  over I @ =
	IF  I' cell- @ I ! dup $@len cell- over $!len  THEN
    cell +LOOP 2drop ;

also robot-voc also definitions player

: N  1 y +! ;
: S -1 y +! ;
: E  1 x +! ;
: W -1 x +! ;
: X get# x ! ;
: Y get# y ! ;
: P get# robot id = IF sp@ cell my-p $+!  THEN  mark-robot ;
: D get# id robot = IF dup my-p delete-p THEN  mark-pos ;
: Robot get# to robot ;
: died. alife off ;

previous previous definitions also robot-voc also Player

: *** cr 0 parse type ;

: #my id to robot ;

: eval-robot ( -- )
    0 to robot alife maxrob cells erase
    pad get-line ( 2dup type cr ) only robot-voc evaluate onlyforth ;

\ flood fill algorithm

: way, ( addr n -- ) over ! buf1 buf1# @ cells + ! 1 buf1# +! ;

: swap-buf ( -- )
    buf1# @ buf2# ! buf1# off buf1 buf2 to buf1 to buf2 ;

\ annotate map with known cost factors

Create costtab here $100 dup allot erase
5   costtab '. + c! \ normal ground is ok
100 costtab ', + c! \ deadly ground is more expensive
200 costtab '; + c! \ corner deadly ground is more expensive
1   costtab '@ + c! \ moves over homes are cheaper
2   costtab '_ + c! \ curious ground is cheaper
10  costtab ': + c! \ difficult ground is more expensive
20  costtab 'e + c! \ enemy ground is more expensive
0   costtab 'E + c! \ cornered enemy ground is not passable

: no-enemies     5 costtab ', + c!   5 costtab '; + c! ;
: have-enemies 100 costtab ', + c! 200 costtab '; + c!
    10 20 random + costtab 'e + c! ;

: moveto? ( n -- flag ) map'' + c@ costtab + c@ ; macro

Variable enemies

: update-enemies ( -- ) enemies off
    maxrob 0 ?DO  I to robot I id <> alife @ and IF
	    1 enemies +!
	    pos 'e over map'' + c!
	    dup 1- moveto? over w - moveto? or 0= IF  'E over map'' + c!  THEN
	    dup 1+ moveto? over w - moveto? or 0= IF  'E over map'' + c!  THEN
	    dup 1- moveto? over w + moveto? or 0= IF  'E over map'' + c!  THEN
	    dup 1+ moveto? over w + moveto? or 0= IF  'E over map'' + c!  THEN
	    drop  THEN
    LOOP  enemies @ IF  have-enemies  ELSE  no-enemies  THEN ;

: update-curious ( -- ) curious-packs $@ bounds ?DO
    '_ I @ map'' + c!  2 cells +LOOP ;

: update-map ( -- ) map' map'' w h * move
    update-enemies update-curious ;

: try, ( addr n -- ) over costs - cell/ moveto?
    dup 0= IF  drop 2drop  EXIT  THEN  4 << +
    over @ -$10 and over -$10 and 2dup
    = IF  2drop over @ or swap !  EXIT  THEN
    u> IF  way,  ELSE  2drop  THEN ;

: try-ewns ( addr n -- ) >r
    dup cell+ r@ try,
    dup cell- r@ try,
    dup w cells + r@ try,
    w cells - r> try, ;

: exec-buf ( -- )
    buf2 buf2# @ cells bounds ?DO
	I @ dup @ try-ewns
    cell +LOOP ;

: init-calc ( x y -- )
    costs w h * cells $FF fill  buf1# off buf2# off
    w * + cells costs + dup $10 way, swap-buf
    dup cell+ $21 try,
    dup cell- $22 try,
    dup w cells + $24 try,
    w cells - $28 try, ;

: calculate-way ( x y -- )  init-calc
    BEGIN  buf1# @  WHILE  swap-buf exec-buf  REPEAT ;

: costs! ( -- ) update-map #my x @ y @ calculate-way ;

\ debugging functions

: .costs ( -- ) base push hex
    costs w h * cells bounds ?DO
	cr I w cells bounds ?DO I @ $FF and 3 u.r cell +LOOP
    w cells +LOOP ;

: .way ( -- ) base push hex
    costs w h * cells bounds ?DO
	cr I w cells bounds ?DO I @ 4 >> $FF and 3 u.r cell +LOOP
    w cells +LOOP ;

: robot-path ( -- ) at?
    maxrob 0 DO  I to robot alife @ IF
	    y @ 1+ x @ at I 0 .r  THEN
    LOOP at ;

: .map2 ( -- ) page .map'' robot-path ;

: .robs ( -- )
    maxrob 0 DO  I to robot alife @ IF  cr '# emit I . .rob  THEN
    LOOP ;

\ Package picking

: pack, ( id destx desty w -- )
    pack cell+ ! w * + pack 2 cells + ! pack !
    pos pack 3 cells + !
    packs $@ bounds ?DO
	I @ pack @ = IF  unloop  EXIT  THEN
	4 cells +LOOP
    pack 4 cells packs $+! ;

: read-packages ( -- )
    #my depth >r pad get-line ( 2dup type cr ) evaluate
    depth r> ?DO  pack,  4 +LOOP ;

: .packs ( -- )  packs $@ bounds
    ?DO  cr I ? '( emit I 2 cells + @ w /mod 0 .r ', emit 0 .r
	." ) " I 3 cells + @ dup 0< IF  negate '# emit 0 .r
	    ELSE  ." @(" w /mod 0 .r ', emit 0 .r ') emit  THEN
	space I cell+ ? 4 cells +LOOP ;

\ find home bases

Variable homes
Variable visited

: home, ( index -- )  sp@ cell homes $+! drop ;

: find-homes ( -- ) s" " homes $!
    map w h * bounds ?DO
	I c@ '@ = IF  I map - home,  THEN
    LOOP ;

: cheapest-home ( -- dir-flag )
    -$10 homes $@ bounds ?DO I @ cells costs + @ umin cell +LOOP $F and ;

: visited-homes ( -- )  pos
    homes $@ bounds ?DO
	dup I @ = IF  I I' cell homes cut-tail  LEAVE
	THEN
    cell +LOOP  drop ;

: visited-curious ( -- )
    pos curious-packs $@ bounds ?DO
	dup I @ = IF
	    cr ." Removing pack " I cell+ @ . ." from curious"
	    I I' 2 cells curious-packs cut-tail  LEAVE  THEN
	2 cells +LOOP  drop ;

: drop-visited ( -- ) visited-homes visited-curious ;

\ deliver packages

: cheapest-package ( -- dir )
    -$10 my-p $@ bounds ?DO
	I @ find-pack pack 2 cells + @ cells costs + @ umin
    cell +LOOP $F and ;

\ compute move out of direction flag

Create moves
,"  " ," E" ," W" ," EW"
," N" ," EN" ," WN" ," EWN"
," S" ," ES" ," WS" ," EWS"
," NS" ," ENS" ," WNS" ," EWNS"

: dir>move ( dir-flag -- char )
    moves swap 0 ?DO count + LOOP
    count random + c@ ;

: hold$ ( addr u -- )  1- FOR  dup I + c@ hold  NEXT  drop ;

: +bid ( n -- n )  money over abs - to money ;

: move" ( dir bid -- addr u ) +bid decimal
    extend rot dup bl = IF  drop <# s" Drop" hold$
    ELSE  <# dir>move hold s" Move " hold$  THEN
    bl hold tuck dabs #S rot sign #> ;

\ Try and pick

: calc-carry ( -- n )
    carry my-p $@ bounds ?DO  I @ find-pack pack cell+ @ -  cell +LOOP ;

: closest-pack ( -- id ) #my -1 calc-carry { pick-pack rest-carry |
    pos -$10  packs $@ bounds ?DO
	over I 3 cells + @ = IF
	    dup I 2 cells + @ cells costs + @ u> I cell+ @ rest-carry u<= and
	    IF  drop I 2 cells + @ cells costs + @ I @ to pick-pack  THEN
	THEN
    4 cells +LOOP 2drop pick-pack } ;

: nearest-package ( -- dir-flag )
    -$10 packs $@ bounds ?DO
	I 3 cells + @ cells costs + @ umin
    4 cells +LOOP $F and ;

: pick-packs ( -- pack1 .. packn n )
    my-p $@len >r
    BEGIN  closest-pack  dup 0>= WHILE  sp@ cell my-p $+! mark-robot  REPEAT
    drop my-p $@ r@ /string dup >r bounds ?DO
	I @ dup unmark-robot
    cell +LOOP  r> cell/ r> my-p $!len ;

: drop-packs ( -- pack1 .. packn n )
    0 #my pos
    my-p $@ bounds ?DO  I @ find-pack dup pack 2 cells + @ =
	IF  I @ -rot >r 1+ r> THEN
    cell +LOOP drop ;

: count-packs ( -- n )
    0 packs $@ bounds ?DO  I 3 cells + @ 0> -  4 cells +LOOP ;

Variable $cratch

: bid# ( n -- )  +bid decimal extend tuck dabs <# #S rot sign #> $cratch $! ;

: pick" ( pack1 .. packn n bid -- addr u ) bid# s"  Pick" $cratch $+!
    0 ?DO  0 <# #S bl hold #> $cratch $+!  LOOP
    $cratch $@ ;

: drop" ( pack1 .. packn n bid -- ) bid# s"  Drop" $cratch $+!
    0 ?DO  0 <# #S bl hold #> $cratch $+!  LOOP
    $cratch $@ ;

\ startup

: round ( -- )
    eval-robot read-packages costs! drop-visited ;
: start ( addr u n -- )
    open-socket to fd
    alife drop x drop y drop
    s" " my-p $!
    s" " other-p $!
    s" " curious-packs $!
    s" " packs $!
    s" Player" fd write-line throw
    read-map read-par find-homes round ;
: command ( addr u -- ) fd write-line throw round ;

\ test individual strategies

: <bid> 2 random 2* 1- ;

: >home
    cheapest-home <bid> move" command
    .map2 .robs .packs ;

: >pack
    cheapest-package <bid> move" command
    .map2 .robs .packs ;

: >pick  pick-packs <bid> pick" command
    .map2 .robs .packs ;

: >drop  drop-packs <bid> drop" command
    .map2 .robs .packs ;

\ Player module

: play ( -- )
    drop-packs ?dup IF  <bid> drop" command  EXIT  THEN
    pick-packs ?dup IF  <bid> pick" command  EXIT  THEN
    my-p $@len  IF  cheapest-package <bid> move" command  EXIT  THEN
    count-packs IF  nearest-package <bid> move" command  EXIT  THEN
    homes $@len IF  cheapest-home <bid> move" command  EXIT  THEN
    cr ." Don't know what to do, killing myself" 0 0 drop" command ;

: plays ( n -- ) .map2 0 ?DO play robot-path
    #my alife @ 0= ?LEAVE  LOOP .robs .packs ;

Variable verbose

main: argc @ 4 = IF 3 arg s" -v" compare 0= verbose !
ELSE  argc @ 3 <> abort" Wrong number of parameters"  THEN
1 arg 2 arg s>number drop start
verbose @ IF .map2 THEN
BEGIN play verbose @ IF robot-path THEN #my alife @ 0= UNTIL bye ;

export plays ;

module;
